# Copyright (C) 2000-2016 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Configure linkchecker using command-line options and configuration.
"""

import codecs
import getpass

from .. import fileutil
from .. import i18n
from .. import logger

from .. import LOG_CMDLINE
from .. import get_link_pat, log

from ..cmdline import print_version, print_usage, print_plugins
from ..director import console


def has_encoding(encoding):
    """Detect if Python can encode in a certain encoding."""
    try:
        codecs.lookup(encoding)
        return True
    except LookupError:
        return False


def setup_config(config, options):
    """Set up linkchecker based on command-line options and configuration"""
    _username = None
    _password = None

    # test if running with -O
    if options.debug and not __debug__:
        log.warn(LOG_CMDLINE, _("Running with python -O disables debugging."))
    # apply commandline options and arguments to configuration
    constructauth = False
    if options.version:
        print_version()
    if not options.warnings:
        config["warnings"] = options.warnings
    if options.externstrict:
        pats = [get_link_pat(arg, strict=True) for arg in options.externstrict]
        config["externlinks"].extend(pats)
    if options.extern:
        pats = [get_link_pat(arg) for arg in options.extern]
        config["externlinks"].extend(pats)
    if options.norobotstxt is not None:
        config["robotstxt"] = options.norobotstxt
    if options.checkextern:
        config["checkextern"] = True
    elif not config["checkextern"]:
        log.info(
            LOG_CMDLINE,
            "Checking intern URLs only; use --check-extern to check extern URLs.",
        )

    if options.output:
        if "/" in options.output:
            logtype, encoding = options.output.split("/", 1)
        else:
            logtype, encoding = options.output, i18n.default_encoding
        logtype = logtype.lower()
        if logtype == "blacklist":
            log.warn(
                LOG_CMDLINE,
                _("blacklist is deprecated for option %(option)s, "
                  "using failures instead") % {"option": "'-o, --output'"}
            )
            logtype = "failures"
        if logtype not in logger.LoggerNames:
            print_usage(
                _("Unknown logger type %(type)r in %(output)r for option %(option)s")
                % {"type": logtype,
                   "output": options.output,
                   "option": "'-o, --output'"}
            )
        if logtype != "none" and not has_encoding(encoding):
            print_usage(
                _("Unknown encoding %(encoding)r in %(output)r for option %(option)s")
                % {
                    "encoding": encoding,
                    "output": options.output,
                    "option": "'-o, --output'",
                }
            )
        config["output"] = logtype
        config["logger"] = config.logger_new(logtype, encoding=encoding)
    if options.fileoutput:
        ns = {"fileoutput": 1}
        for arg in options.fileoutput:
            ftype = arg
            # look for (optional) filename and encoding
            if "/" in ftype:
                ftype, suffix = ftype.split("/", 1)
                if suffix:
                    if has_encoding(suffix):
                        # it was an encoding
                        ns["encoding"] = suffix
                    elif "/" in suffix:
                        # look for (optional) encoding
                        encoding, filename = suffix.split("/", 1)
                        if has_encoding(encoding):
                            ns["encoding"] = encoding
                            ns["filename"] = filename
                        else:
                            ns["filename"] = suffix
                    else:
                        ns["filename"] = suffix
            if ftype == "blacklist":
                log.warn(
                    LOG_CMDLINE,
                    _("blacklist logger is deprecated for option %(option)s, "
                      "using failures instead") % {"option": "'-F, --file-output'"}
                )
                ftype = "failures"
            if ftype not in logger.LoggerNames:
                print_usage(
                    _("Unknown logger type %(type)r in %(output)r"
                      " for option %(option)s")
                    % {
                        "type": ftype,
                        "output": options.fileoutput,
                        "option": "'-F, --file-output'",
                    }
                )
            if ftype != "none" and "encoding" in ns \
                    and not has_encoding(ns["encoding"]):
                print_usage(
                    _("Unknown encoding %(encoding)r in %(output)r"
                      " for option %(option)s")
                    % {
                        "encoding": ns["encoding"],
                        "output": options.fileoutput,
                        "option": "'-F, --file-output'",
                    }
                )
            new_logger = config.logger_new(ftype, **ns)
            config["fileoutput"].append(new_logger)
    if options.nntpserver:
        config["nntpserver"] = options.nntpserver
    if options.username:
        _username = options.username
        constructauth = True
    if options.password:
        if _username:
            msg = _("Enter LinkChecker HTTP/FTP password for user %(user)s:") % {
                "user": _username
            }
        else:
            msg = _("Enter LinkChecker HTTP/FTP password:")
        _password = getpass.getpass(console.encode(msg))
        constructauth = True
    if options.quiet:
        config["logger"] = config.logger_new("none")
    if options.recursionlevel is not None:
        config["recursionlevel"] = options.recursionlevel
    if options.status is not None:
        config["status"] = options.status
    if options.threads is not None:
        if options.threads < 1:
            options.threads = 0
        config["threads"] = options.threads
    if options.timeout is not None:
        if options.timeout > 0:
            config["timeout"] = options.timeout
        else:
            print_usage(
                _("Illegal argument %(arg)r for option %(option)s")
                % {"arg": options.timeout, "option": "'--timeout'"}
            )
    if options.listplugins:
        print_plugins(config["pluginfolders"])
    if options.verbose:
        if options.verbose:
            config["verbose"] = True
            config["warnings"] = True
    if options.cookiefile is not None:
        config["cookiefile"] = options.cookiefile
    if constructauth:
        config.add_auth(pattern=".+", user=_username, password=_password)
    # read missing passwords
    for entry in config["authentication"]:
        if entry["password"] is None:
            attrs = entry.copy()
            attrs["strpattern"] = attrs["pattern"].pattern
            msg = (
                _("Enter LinkChecker password for user %(user)s at %(strpattern)s:")
                % attrs
            )
            entry["password"] = getpass.getpass(msg)
    if options.useragent is not None:
        config["useragent"] = options.useragent
    if options.cookiefile is not None:
        if fileutil.is_readable(options.cookiefile):
            config["cookiefile"] = options.cookiefile
        else:
            msg = _("Could not read cookie file %s") % options.cookiefile
            log.error(LOG_CMDLINE, msg)



# Copyright (C) 2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
HTML form utils
"""
from ..htmlutil import htmlsoup
from .. import log, LOG_CHECK


class Form:
    """Store HTML form URL and form data."""

    def __init__(self, url):
        """Set URL and empty form data."""
        self.url = url
        self.data = {}

    def add_value(self, key, value):
        """Add a form value."""
        self.data[key] = value

    def __repr__(self):
        """Return string displaying URL and form data."""
        return f"<url={self.url} data={self.data}>"


def search_form(content, cgiuser, cgipassword):
    """Search for a HTML form in the given HTML content that has input elements
    with name attributes that match cgiuser and/or cgipassword. If no such form
    is found return None.
    """
    soup = htmlsoup.make_soup(content)
    cginames = {cgiuser, cgipassword} - {None}
    for form_element in soup.find_all("form", action=True):
        form = Form(form_element["action"])
        for input_element in form_element.find_all("input", attrs={"name": True}):
            form.add_value(input_element["name"], input_element.attrs.get("value"))
        if cginames <= set(form.data):
            log.debug(LOG_CHECK, "Found form %s", form)
            return form

    # not found
    log.warn(LOG_CHECK, "Form with fields %s not found", ",".join(cginames))
    return None



# Copyright (C) 2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Cache check results.
"""
from ..decorators import synchronized
from ..lock import get_lock


# lock object
cache_lock = get_lock("results_cache_lock")


class ResultCache:
    """
    Thread-safe cache of UrlData.to_wire() results.
    the cache is limited in size since we rather recheck the same URL
    multiple times instead of running out of memory.
    format: {cache key (string) -> result (UrlData.towire())}
    """

    def __init__(self, result_cache_size):
        """Initialize result cache."""
        # mapping {URL -> cached result}
        self.cache = {}
        self.max_size = result_cache_size

    @synchronized(cache_lock)
    def get_result(self, key):
        """Return cached result or None if not found."""
        return self.cache.get(key)

    @synchronized(cache_lock)
    def add_result(self, key, result):
        """Add result object to cache with given key.
        The request is ignored when the cache is already full or the key
        is None.
        """
        if len(self.cache) > self.max_size:
            return
        if key is not None:
            self.cache[key] = result

    def has_result(self, key):
        """Non-thread-safe function for fast containment checks."""
        return key in self.cache

    def has_non_empty_result(self, key):
        """Non-thread-safe function for fast containment checks."""
        return self.cache.get(key)

    def __len__(self):
        """Get number of cached elements. This is not thread-safe and is
        likely to change before the returned value is used."""
        return len(self.cache)



# Copyright (C) 2006-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Status message handling"""
import time
from . import task


class Status(task.LoggedCheckedTask):
    """Thread that gathers and logs the status periodically."""

    def __init__(self, aggregator, wait_seconds):
        """Initialize the status logger task.

        @param urlqueue: the URL queue
        @type urlqueue: Urlqueue
        @param logger: the logger object to inform about status
        @type logger: console.StatusLogger
        @param wait_seconds: interval in seconds to report status
        @type wait_seconds: int
        """
        logger = aggregator.config.status_logger
        super().__init__(logger)
        self.aggregator = aggregator
        self.wait_seconds = wait_seconds
        assert self.wait_seconds >= 1

    def run_checked(self):
        """Print periodic status messages."""
        self.start_time = time.time()
        self.name = "Status"
        # the first status should be after a second
        wait_seconds = 1
        first_wait = True
        while not self.stopped(wait_seconds):
            self.log_status()
            if first_wait:
                wait_seconds = self.wait_seconds
                first_wait = False

    def log_status(self):
        """Log a status message."""
        duration = time.time() - self.start_time
        checked, in_progress, queue = self.aggregator.urlqueue.status()
        num_urls = len(self.aggregator.result_cache)
        self.logger.log_status(checked, in_progress, queue, duration, num_urls)



__app_name__ = "LinkChecker"
__version__ = "10.2.1"
__release_date__ = "2022-12-05"
__copyright_year__ = "2022"
__author__ = "LinkChecker Authors"
__url__ = "https://linkchecker.github.io/linkchecker/"
__support_url__ = "https://github.com/linkchecker/linkchecker/issues"



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Check page content with regular expression.
"""
import re
from . import _ContentPlugin
from .. import log, LOG_PLUGIN


class RegexCheck(_ContentPlugin):
    """Define a regular expression which prints a warning if it matches
    any content of the checked link. This applies only to valid pages,
    so we can get their content.

    Use this to check for pages that contain some form of error
    message, for example 'This page has moved' or 'Oracle
    Application error'.

    Note that multiple values can be combined in the regular expression,
    for example "(This page has moved|Oracle Application error)"."""

    def __init__(self, config):
        """Set warning regex from config."""
        super().__init__(config)
        self.warningregex = None
        pattern = config["warningregex"]
        if pattern:
            try:
                self.warningregex = re.compile(pattern)
            except re.error as msg:
                log.warn(LOG_PLUGIN, _("Invalid regex pattern %r: %s"), pattern, msg)

    def applies_to(self, url_data):
        """Check for warningregex, extern flag and parseability."""
        return self.warningregex and not url_data.extern[0] and url_data.is_parseable()

    def check(self, url_data):
        """Check content."""
        log.debug(LOG_PLUGIN, "checking content for warning regex")
        content = url_data.get_content()
        # add warnings for found matches, up to the maximum allowed number
        match = self.warningregex.search(content)
        if match:
            # calculate line number for match
            line = content.count('\n', 0, match.start())
            # add a warning message
            msg = _("Found %(match)r at line %(line)d in link contents.")
            url_data.add_warning(msg % {"match": match.group(), "line": line})

    @classmethod
    def read_config(cls, configparser):
        """Read configuration file options."""
        config = dict()
        section = cls.__name__
        option = "warningregex"
        if configparser.has_option(section, option):
            value = configparser.get(section, option)
        else:
            value = None
        config[option] = value
        return config



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Main functions for link checking.
"""

import os
import html
import urllib.parse

from .. import url as urlutil, log, LOG_CHECK

MAX_FILESIZE = 1024 * 1024 * 10  # 10MB


def guess_url(url):
    """Guess if URL is a http or ftp URL.

    @param url: the URL to check
    @type url: unicode
    @return: url with *http://* or *ftp://* prepended if it's detected as
        a http respective ftp URL.
    @rtype: unicode
    """
    if url.lower().endswith(".html") and "/" not in url:
        return url
    if url.lower().startswith("www."):
        # syntactic sugar
        return f"http://{url}"
    elif url.lower().startswith("ftp."):
        # syntactic sugar
        return f"ftp://{url}"
    return url


def absolute_url(base_url, base_ref, parent_url):
    """
    Search for the absolute url to detect the link type. This does not
    join any url fragments together!

    @param base_url: base url from a link tag
    @type base_url: string or None
    @param base_ref: base url from <base> tag
    @type base_ref: string or None
    @param parent_url: url of parent document
    @type parent_url: string or None
    """
    if base_url and urlutil.url_is_absolute(base_url):
        return base_url
    elif base_ref and urlutil.url_is_absolute(base_ref):
        return base_ref
    elif parent_url and urlutil.url_is_absolute(parent_url):
        return parent_url
    return ""


def get_url_from(
    base_url,
    recursion_level,
    aggregate,
    parent_url=None,
    base_ref=None,
    line=None,
    column=None,
    page=0,
    name="",
    parent_content_type=None,
    extern=None,
    url_encoding=None,
):
    """
    Get url data from given base data.

    @param base_url: base url from a link tag
    @type base_url: string or None
    @param recursion_level: current recursion level
    @type recursion_level: number
    @param aggregate: aggregate object
    @type aggregate: aggregate.Consumer
    @param parent_url: parent url
    @type parent_url: string or None
    @param base_ref: base url from <base> tag
    @type base_ref: string or None
    @param line: line number
    @type line: number
    @param column: column number
    @type column: number
    @param page: page number
    @type page: number
    @param name: link name
    @type name: string
    @param extern: (is_extern, is_strict) or None
    @type extern: tuple(int, int) or None
    """
    if base_url is not None:
        # left strip for detection of URL scheme
        base_url_stripped = base_url.lstrip()
    else:
        base_url_stripped = base_url
    url = absolute_url(base_url_stripped, base_ref, parent_url).lower()
    if ":" in url:
        scheme = url.split(":", 1)[0].lower()
    else:
        scheme = None
        if not (url or name):
            # use filename as base url, with slash as path separator
            name = base_url.replace("\\", "/")
    allowed_schemes = aggregate.config["allowedschemes"]
    # ignore local PHP files with execution directives
    local_php = (
        parent_content_type == 'application/x-httpd-php'
        and '<?' in base_url
        and '?>' in base_url
        and scheme == 'file'
    )
    if local_php or (allowed_schemes and scheme not in allowed_schemes):
        klass = ignoreurl.IgnoreUrl
    else:
        assume_local_file = recursion_level == 0
        klass = get_urlclass_from(scheme, assume_local_file=assume_local_file)
        if "AnchorCheck" in aggregate.config["enabledplugins"] and \
                klass == fileurl.FileUrl:
            klass = fileurl.AnchorCheckFileUrl
    log.debug(LOG_CHECK, "%s handles url %s", klass.__name__, base_url)
    return klass(
        base_url,
        recursion_level,
        aggregate,
        parent_url=parent_url,
        base_ref=base_ref,
        line=line,
        column=column,
        page=page,
        name=name,
        extern=extern,
        url_encoding=url_encoding,
    )


def get_urlclass_from(scheme, assume_local_file=False):
    """Return checker class for given URL scheme. If the scheme
    cannot be matched and assume_local_file is True, assume a local file.
    """
    if scheme in ("http", "https"):
        klass = httpurl.HttpUrl
    elif scheme == "ftp":
        klass = ftpurl.FtpUrl
    elif scheme == "file":
        klass = fileurl.FileUrl
    elif scheme == "telnet":
        klass = telneturl.TelnetUrl
    elif scheme == "mailto":
        klass = mailtourl.MailtoUrl
    elif scheme in ("nntp", "news", "snews"):
        klass = nntpurl.NntpUrl
    elif scheme == "dns":
        klass = dnsurl.DnsUrl
    elif scheme == "itms-services":
        klass = itmsservicesurl.ItmsServicesUrl
    elif scheme and unknownurl.is_unknown_scheme(scheme):
        klass = unknownurl.UnknownUrl
    elif assume_local_file:
        klass = fileurl.FileUrl
    else:
        klass = unknownurl.UnknownUrl
    return klass


def get_index_html(urls):
    """
    Construct artificial index.html from given URLs.

    @param urls: URL strings
    @type urls: iterator of string
    """
    lines = ["<html>", "<body>"]
    for entry in urls:
        name = html.escape(entry)
        try:
            url = html.escape(urllib.parse.quote(entry))
        except UnicodeEncodeError:
            log.warn(LOG_CHECK, "Unable to convert entry to Unicode")
            continue
        except KeyError:
            # Some unicode entries raise KeyError.
            url = name
        lines.append(f'<a href="{url}">{name}</a>')
    lines.extend(["</body>", "</html>"])
    return os.linesep.join(lines)


# all the URL classes
from . import (  # noqa: E402
    fileurl,
    unknownurl,
    ftpurl,
    httpurl,
    dnsurl,
    mailtourl,
    telneturl,
    nntpurl,
    ignoreurl,
    itmsservicesurl,
)



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
The default text logger.
"""
import time

from . import _Logger
from .. import ansicolor, strformat, configuration, i18n


class TextLogger(_Logger):
    """
    A text logger, colorizing the output if possible.

    Informal text output format spec:
    Output consists of a set of URL logs separated by one or more
    blank lines.
    A URL log consists of two or more lines. Each line consists of
    keyword and data, separated by whitespace.
    Unknown keywords will be ignored.
    """

    LoggerName = 'text'

    LoggerArgs = {
        "filename": "linkchecker-out.txt",
        'colorparent': "default",
        'colorurl': "default",
        'colorname': "default",
        'colorreal': "cyan",
        'colorbase': "purple",
        'colorvalid': "bold;green",
        'colorinvalid': "bold;red",
        'colorinfo': "default",
        'colorwarning': "bold;yellow",
        'colordltime': "default",
        'colordlsize': "default",
        'colorreset': "default",
    }

    def __init__(self, **kwargs):
        """Initialize error counter and optional file output."""
        args = self.get_args(kwargs)
        super().__init__(**args)
        self.output_encoding = args.get("encoding", i18n.default_encoding)
        self.init_fileoutput(args)
        self.colorparent = args.get('colorparent', 'default')
        self.colorurl = args.get('colorurl', 'default')
        self.colorname = args.get('colorname', 'default')
        self.colorreal = args.get('colorreal', 'default')
        self.colorbase = args.get('colorbase', 'default')
        self.colorvalid = args.get('colorvalid', 'default')
        self.colorinvalid = args.get('colorinvalid', 'default')
        self.colorinfo = args.get('colorinfo', 'default')
        self.colorwarning = args.get('colorwarning', 'default')
        self.colordltime = args.get('colordltime', 'default')
        self.colordlsize = args.get('colordlsize', 'default')
        self.colorreset = args.get('colorreset', 'default')

    def init_fileoutput(self, args):
        """Colorize file output if possible."""
        super().init_fileoutput(args)
        if self.fd is not None:
            self.fd = ansicolor.Colorizer(self.fd)

    def start_fileoutput(self):
        """Needed to make file descriptor color aware."""
        init_color = self.fd is None
        super().start_fileoutput()
        if init_color:
            self.fd = ansicolor.Colorizer(self.fd)

    def start_output(self):
        """Write generic start checking info."""
        super().start_output()
        if self.has_part('intro'):
            self.write_intro()
        self.flush()

    def write_intro(self):
        """Log introduction text."""
        self.writeln(configuration.App)
        self.writeln(configuration.Copyright)
        self.writeln(configuration.Freeware)
        self.writeln(
            _("Read the documentation at %(url)s") % {'url': configuration.Url}
        )
        self.writeln(
            _("Write comments and bugs to %(url)s") % {'url': configuration.SupportUrl}
        )
        self.check_date()
        self.writeln()
        self.writeln(_("Start checking at %s") % strformat.strtime(self.starttime))

    def log_url(self, url_data):
        """Write url checking info."""
        self.writeln()
        if self.has_part('url'):
            self.write_url(url_data)
        if url_data.name and self.has_part('name'):
            self.write_name(url_data)
        if url_data.parent_url and self.has_part('parenturl'):
            self.write_parent(url_data)
        if url_data.base_ref and self.has_part('base'):
            self.write_base(url_data)
        if url_data.url and self.has_part('realurl'):
            self.write_real(url_data)
        if url_data.checktime and self.has_part('checktime'):
            self.write_checktime(url_data)
        if url_data.dltime >= 0 and self.has_part('dltime'):
            self.write_dltime(url_data)
        if url_data.size >= 0 and self.has_part('dlsize'):
            self.write_size(url_data)
        if url_data.info and self.has_part('info'):
            self.write_info(url_data)
        if url_data.modified and self.has_part('modified'):
            self.write_modified(url_data)
        if url_data.warnings and self.has_part('warning'):
            self.write_warning(url_data)
        if self.has_part('result'):
            self.write_result(url_data)
        self.flush()

    def write_id(self):
        """Write unique ID of url_data."""
        self.writeln()
        self.write(self.part('id') + self.spaces('id'))
        self.writeln("%d" % self.stats.number, color=self.colorinfo)

    def write_url(self, url_data):
        """Write url_data.base_url."""
        self.write(self.part('url') + self.spaces('url'))
        txt = strformat.strline(url_data.base_url)
        self.writeln(txt, color=self.colorurl)

    def write_name(self, url_data):
        """Write url_data.name."""
        self.write(self.part("name") + self.spaces("name"))
        self.writeln(strformat.strline(url_data.name), color=self.colorname)

    def write_parent(self, url_data):
        """Write url_data.parent_url."""
        self.write(self.part('parenturl') + self.spaces("parenturl"))
        txt = url_data.parent_url
        if url_data.line is not None:
            txt += _(", line %d") % url_data.line
        if url_data.column is not None:
            txt += _(", col %d") % url_data.column
        if url_data.page > 0:
            txt += _(", page %d") % url_data.page
        self.writeln(txt, color=self.colorparent)

    def write_base(self, url_data):
        """Write url_data.base_ref."""
        self.write(self.part("base") + self.spaces("base"))
        self.writeln(url_data.base_ref, color=self.colorbase)

    def write_real(self, url_data):
        """Write url_data.url."""
        self.write(self.part("realurl") + self.spaces("realurl"))
        self.writeln(url_data.url, color=self.colorreal)

    def write_dltime(self, url_data):
        """Write url_data.dltime."""
        self.write(self.part("dltime") + self.spaces("dltime"))
        self.writeln(_("%.3f seconds") % url_data.dltime, color=self.colordltime)

    def write_size(self, url_data):
        """Write url_data.size."""
        self.write(self.part("dlsize") + self.spaces("dlsize"))
        self.writeln(strformat.strsize(url_data.size), color=self.colordlsize)

    def write_checktime(self, url_data):
        """Write url_data.checktime."""
        self.write(self.part("checktime") + self.spaces("checktime"))
        self.writeln(_("%.3f seconds") % url_data.checktime, color=self.colordltime)

    def write_info(self, url_data):
        """Write url_data.info."""
        self.write(self.part("info") + self.spaces("info"))
        self.writeln(self.wrap(url_data.info, 65), color=self.colorinfo)

    def write_modified(self, url_data):
        """Write url_data.modified."""
        self.write(self.part("modified") + self.spaces("modified"))
        self.writeln(self.format_modified(url_data.modified))

    def write_warning(self, url_data):
        """Write url_data.warning."""
        self.write(self.part("warning") + self.spaces("warning"))
        warning_msgs = [f"[{tag}] {msg}" for tag, msg in url_data.warnings]
        self.writeln(self.wrap(warning_msgs, 65), color=self.colorwarning)

    def write_result(self, url_data):
        """Write url_data.result."""
        self.write(self.part("result") + self.spaces("result"))
        if url_data.valid:
            color = self.colorvalid
            self.write(_("Valid"), color=color)
        else:
            color = self.colorinvalid
            self.write(_("Error"), color=color)
        if url_data.result:
            self.write(": " + url_data.result, color=color)
        self.writeln()

    def write_outro(self, interrupt=False):
        """Write end of checking message."""
        self.writeln()
        if interrupt:
            self.writeln(_("The check has been interrupted; results are not complete."))
        self.write(_("That's it.") + " ")
        self.write(_n("%d link", "%d links", self.stats.number) % self.stats.number)
        self.write(" ")
        if self.stats.num_urls is not None:
            self.write(
                _n("in %d URL", "in %d URLs", self.stats.num_urls) % self.stats.num_urls
            )
        self.write(_(" checked.") + " ")
        warning_text = (
            _n("%d warning found", "%d warnings found", self.stats.warnings_printed)
            % self.stats.warnings_printed
        )
        if self.stats.warnings_printed:
            warning_color = self.colorwarning
        else:
            warning_color = self.colorinfo
        self.write(warning_text, color=warning_color)
        if self.stats.warnings != self.stats.warnings_printed:
            self.write(
                _(" (%d ignored or duplicates not printed)")
                % (self.stats.warnings - self.stats.warnings_printed)
            )
        self.write(". ")
        error_text = (
            _n("%d error found", "%d errors found", self.stats.errors_printed)
            % self.stats.errors_printed
        )
        if self.stats.errors_printed:
            error_color = self.colorinvalid
        else:
            error_color = self.colorvalid
        self.write(error_text, color=error_color)
        if self.stats.errors != self.stats.errors_printed:
            self.write(
                _(" (%d duplicates not printed)")
                % (self.stats.errors - self.stats.errors_printed)
            )
        self.writeln(".")
        num = self.stats.internal_errors
        if num:
            self.writeln(
                _n(
                    "There was %(num)d internal error.",
                    "There were %(num)d internal errors.",
                    num,
                )
                % {"num": num}
            )
        self.stoptime = time.time()
        duration = self.stoptime - self.starttime
        self.writeln(
            _("Stopped checking at %(time)s (%(duration)s)")
            % {
                "time": strformat.strtime(self.stoptime),
                "duration": strformat.strduration_long(duration),
            }
        )

    def write_stats(self):
        """Write check statistic info."""
        self.writeln()
        self.writeln(_("Statistics:"))
        if self.stats.downloaded_bytes is not None:
            self.writeln(
                _("Downloaded: %s.") % strformat.strsize(self.stats.downloaded_bytes)
            )
        if self.stats.number > 0:
            self.writeln(
                _(
                    "Content types: %(image)d image, %(text)d text, %(video)d video, "
                    "%(audio)d audio, %(application)d application, %(mail)d mail"
                    " and %(other)d other."
                )
                % self.stats.link_types
            )
            self.writeln(
                _("URL lengths: min=%(min)d, max=%(max)d, avg=%(avg)d.")
                % dict(
                    min=self.stats.min_url_length,
                    max=self.stats.max_url_length,
                    avg=self.stats.avg_url_length,
                )
            )
        else:
            self.writeln(_("No statistics available since no URLs were checked."))

    def end_output(self, **kwargs):
        """Write end of output info, and flush all output buffers."""
        self.stats.downloaded_bytes = kwargs.get("downloaded_bytes")
        self.stats.num_urls = kwargs.get("num_urls")
        if self.has_part('stats'):
            self.write_stats()
        if self.has_part('outro'):
            self.write_outro(interrupt=kwargs.get("interrupt"))
        self.close_fileoutput()



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
import threading
import time
import requests
from xml.dom.minidom import parseString
from . import _ContentPlugin
from .. import log, LOG_PLUGIN
from ..decorators import synchronized


_w3_time_lock = threading.Lock()


class W3Timer:
    """Ensure W3C apis are not hammered."""

    # every X seconds
    SleepSeconds = 2

    def __init__(self):
        """Remember last API call."""
        self.last_w3_call = 0

    @synchronized(_w3_time_lock)
    def check_w3_time(self):
        """Make sure the W3C validators are at most called once a second."""
        if time.time() - self.last_w3_call < W3Timer.SleepSeconds:
            time.sleep(W3Timer.SleepSeconds)
        self.last_w3_call = time.time()


class HtmlSyntaxCheck(_ContentPlugin):
    """Check the syntax of HTML pages with the online W3C HTML validator.
    See https://validator.w3.org/docs/api.html.
    """

    def __init__(self, config):
        """Initialize plugin."""
        super().__init__(config)
        self.timer = W3Timer()
        log.warn(
            LOG_PLUGIN, _("HTML syntax check plugin is broken. Fixes welcome.")
        )

    def applies_to(self, url_data):
        """Check for HTML, extern and session."""
        return False  # XXX Plugin disabled
        return (url_data.is_html() and not url_data.extern[0]
                and hasattr(url_data, "session"))

    def check(self, url_data):
        """Check HTML syntax of given URL."""
        self.timer.check_w3_time()
        session = url_data.session
        try:
            body = {'uri': url_data.url, 'output': 'soap12'}
            response = session.post('https://validator.w3.org/check', data=body)
            response.raise_for_status()
            if response.headers.get('x-w3c-validator-status', 'Invalid') == 'Valid':
                url_data.add_info("W3C Validator: %s" % _("valid HTML syntax"))
                return
            check_w3_errors(url_data, response.text, "W3C HTML")
        except requests.exceptions.RequestException:
            pass  # ignore service failures
        except Exception as msg:
            log.warn(
                LOG_PLUGIN, _("HTML syntax check plugin error: %(msg)s ") % {"msg": msg}
            )


class CssSyntaxCheck(_ContentPlugin):
    """Check the syntax of HTML pages with the online W3C CSS validator.
    See https://jigsaw.w3.org/css-validator/manual.html#expert.
    """

    def __init__(self, config):
        """Initialize plugin."""
        super().__init__(config)
        self.timer = W3Timer()

    def applies_to(self, url_data):
        """Check for CSS, extern and session."""
        return (url_data.is_css() and not url_data.extern[0]
                and hasattr(url_data, "session"))

    def check(self, url_data):
        """Check CSS syntax of given URL."""
        self.timer.check_w3_time()
        session = url_data.session
        try:
            url = 'https://jigsaw.w3.org/css-validator/validator'
            params = {
                'uri': url_data.url,
                'warning': '2',
                'output': 'soap12',
            }
            response = session.get(url, params=params)
            response.raise_for_status()
            if response.headers.get('X-W3C-Validator-Status', 'Invalid') == 'Valid':
                url_data.add_info("W3C Validator: %s" % _("valid CSS syntax"))
                return
            check_w3_errors(url_data, response.text, "W3C CSS")
        except requests.exceptions.RequestException:
            pass  # ignore service failures
        except Exception as msg:
            log.warn(
                LOG_PLUGIN, _("CSS syntax check plugin error: %(msg)s ") % {"msg": msg}
            )


def check_w3_errors(url_data, xml, w3type):
    """Add warnings for W3C HTML or CSS errors in xml format.
    w3type is either "W3C HTML" or "W3C CSS"."""
    dom = parseString(xml)
    for error in dom.getElementsByTagName('m:error'):
        if w3type == "W3C HTML":
            warnmsg = _(
                "%(w3type)s validation error at line %(line)s col %(column)s: %(msg)s"
            )
        else:
            warnmsg = _(
                "%(w3type)s validation error at line %(line)s: %(msg)s"
            )
        attrs = {
            "w3type": w3type,
            "line": getXmlText(error, "m:line"),
            "msg": getXmlText(error, "m:message"),
        }
        if w3type == "W3C HTML":
            attrs["column"] = getXmlText(error, "m:col")
        url_data.add_warning(warnmsg % attrs)


def getXmlText(parent, tag):
    """Return XML content of given tag in parent element."""
    elem = parent.getElementsByTagName(tag)[0]
    # Yes, the DOM standard is awful.
    rc = []
    for node in elem.childNodes:
        if node.nodeType == node.TEXT_NODE:
            rc.append(node.data)
    return ''.join(rc)
